import os
from pathlib import Path
from typing import Optional

import pandas as pd
import torch
from GANDLF.utils import get_unique_timestamp
import lightning.pytorch as pl
from GANDLF.models.lightning_module import GandlfLightningModule
from GANDLF.data.lightning_datamodule import GandlfInferenceDatamodule


def InferenceManager(
    dataframe: pd.DataFrame,
    modelDir: str,
    parameters: dict,
    outputDir: Optional[str] = None,
) -> None:
    """
    This function is used to perform inference on a model using a dataframe.

    Args:
        dataframe (pd.DataFrame): The dataframe containing the data to be used for inference.
        modelDir (str): The path to the model directory.
        parameters (dict): The parameters to be used for inference.
        device (str): The device to be used for inference.
        outputDir (Optional[str], optional): The output directory for the inference results. Defaults to None.
    """
    if outputDir is None:
        outputDir = os.path.join(modelDir, get_unique_timestamp(), "output_inference")
        print(
            "Output directory not provided, creating a new directory with a unique timestamp: ",
            outputDir,
        )
    Path(outputDir).mkdir(parents=True, exist_ok=True)

    parameters["output_dir"] = outputDir

    # initialize parameters for inference
    for key in ["penalty_weights", "sampling_weights", "class_weights"]:
        parameters.setdefault(key, None)

    n_folds = parameters["nested_training"]["validation"]
    modelDir_split = modelDir.split(",") if "," in modelDir else [modelDir]

    # This should be handled by config parser
    accelerator = parameters.get("accelerator", "auto")
    allowed_accelerators = ["cpu", "gpu", "auto"]
    # codacy ignore Generic/ReDoS: This is not a SQL query, it's an error message.
    assert (
        accelerator in allowed_accelerators
    ), f"Invalid accelerator selected: {accelerator}. Please select from {allowed_accelerators}"
    strategy = parameters.get("strategy", "auto")
    allowed_strategies = ["auto", "ddp"]
    # codacy ignore Generic/ReDoS: This is not a SQL query, it's an error message.
    assert (
        strategy in allowed_strategies
    ), f"Invalid strategy selected: {strategy}. Please select from {allowed_strategies}"
    precision = parameters.get("precision", "32")
    allowed_precisions = [
        "64",
        "64-true",
        "32",
        "32-true",
        "16",
        "16-mixed",
        "bf16",
        "bf16-mixed",
    ]
    # codacy ignore Generic/ReDoS: This is not a SQL query, it's an error message.
    assert (
        precision in allowed_precisions
    ), f"Invalid precision selected: {precision}. Please select from {allowed_precisions}"

    averaged_probs_list = []
    for current_modelDir in modelDir_split:
        fold_dirs = (
            [
                os.path.join(current_modelDir, d, "")
                for d in sorted(os.listdir(current_modelDir))
                if d.isdigit()
            ]
            if n_folds > 1
            else [current_modelDir]
        )

        is_classification = parameters["problem_type"] == "classification"
        parameters["model"].setdefault("type", "torch")
        class_list = (
            [str(c) for c in parameters["model"]["class_list"]]
            if is_classification
            else None
        )
        probs_list = None
        for fold_dir in fold_dirs:
            trainer = pl.Trainer(
                accelerator=accelerator,
                strategy=strategy,
                fast_dev_run=False,
                devices=parameters.get("devices", "auto"),
                num_nodes=parameters.get("num_nodes", 1),
                precision=precision,
                gradient_clip_algorithm=parameters["clip_mode"],
                gradient_clip_val=parameters["clip_grad"],
                max_epochs=parameters["num_epochs"],
                sync_batchnorm=False,
                enable_checkpointing=False,
                logger=False,
                num_sanity_val_steps=0,
            )
            datamodule = GandlfInferenceDatamodule(dataframe, parameters)
            parameters = datamodule.updated_parameters_dict
            lightning_module = GandlfLightningModule(parameters, output_dir=fold_dir)

            if parameters.get("auto_batch_size_find", False):
                print(
                    "Auto batch size find is not supported in inference. Dataloader batch size is always 1."
                )

            trainer.predict(lightning_module, datamodule=datamodule)
            if is_classification:
                prob_values_for_all_subjects_in_fold = list(
                    lightning_module.subject_classification_class_probabilities.values()
                )
                if prob_values_for_all_subjects_in_fold:
                    probs_list = torch.stack(
                        prob_values_for_all_subjects_in_fold, dim=1
                    )

        if is_classification and probs_list is not None:
            averaged_probs_list.append(torch.mean(probs_list, 0))

    # this logic should be changed if we want to do multi-fold inference for histo images
    if averaged_probs_list and is_classification:
        averaged_probs_df = pd.DataFrame(
            columns=["SubjectID", "PredictedClass"] + class_list
        )
        averaged_probs_df["SubjectID"] = dataframe.iloc[:, 0]

        averaged_probs_across_models = (
            torch.mean(torch.stack(averaged_probs_list), 0).cpu().numpy()
        )
        averaged_probs_df[class_list] = averaged_probs_across_models
        averaged_probs_df["PredictedClass"] = [
            class_list[idx] for idx in averaged_probs_across_models.argmax(axis=1)
        ]

        filepath_to_save = os.path.join(outputDir, "final_preds_and_avg_probs.csv")
        if os.path.isfile(filepath_to_save):
            filepath_to_save = os.path.join(
                outputDir,
                "final_preds_and_avg_probs_" + get_unique_timestamp() + ".csv",
            )

        averaged_probs_df.to_csv(filepath_to_save, index=False)



# -*- coding: utf-8 -*-
"""
Add Conv Block for MSDNet
"""


def add_conv_block(
    Conv,
    BatchNorm,
    in_channels=1,
    out_channels=1,
    kernel_size=3,
    dilation=1,
    last=False,
):
    """
    Adds a convolution block to a neural network. A convolution block consists of a convolutional layer and a batch
    normalization layer.

    Args:
        Conv (class): Convolutional layer class.
        BatchNorm (class): Batch normalization layer class.
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        kernel_size (int, optional): Kernel size of the convolutional layer. Default is 3.
        dilation (int, optional): Dilation rate of the convolutional layer. Default is 1.
        last (bool, optional): Indicates if this is the last block in the neural network. Default is False.

    Returns:
        list :A list containing the convolutional layer and batch normalization layer.

    """
    padding = dilation if not last else 0
    conv_layer = Conv(
        in_channels, out_channels, kernel_size, padding=padding, dilation=dilation
    )
    bn_layer = BatchNorm(out_channels)

    return [conv_layer, bn_layer]



#!usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
from typing import Optional

import click
from deprecated import deprecated

from GANDLF.cli import copyrightMessage, recover_config
from GANDLF.entrypoints import append_copyright_to_help
from GANDLF.utils import logger_setup


def _recover_config(model_dir: Optional[str], mlcube: bool, output_file: str):
    if mlcube:
        search_dir = "/embedded_model/"
    else:
        search_dir = model_dir

    print(f"{model_dir=}")
    print(f"{mlcube=}")
    print(f"{search_dir=}")
    print(f"{output_file=}")
    result = recover_config(search_dir, output_file)
    assert result, "Config file recovery failed."


@click.command()
@click.option(
    "--model-dir",
    "-m",
    help="Path to the model directory.",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
)
@click.option(
    "--mlcube",
    "-c",
    is_flag=True,
    help="Pass this option to attempt to extract the config from the embedded model in a GaNDLF MLCube "
    "(if any). Only useful in that context. If passed, model-dir param is ignored.",
)
@click.option(
    "--output-file",
    "-o",
    required=True,
    type=click.Path(file_okay=True, dir_okay=False),
    help="Path to an output file where the config will be written.",
)
@click.option(
    "--log-file",
    type=click.Path(),
    default=None,
    help="Output file which will contain the logs.",
)
@append_copyright_to_help
def new_way(model_dir, mlcube, output_file, log_file):
    """Recovers a config file from a GaNDLF model. If used from within a deployed GaNDLF MLCube,
    attempts to extract the config from the embedded model."""

    logger_setup(log_file)
    _recover_config(model_dir=model_dir, mlcube=mlcube, output_file=output_file)


# old-fashioned way of running gandlf via `gandlf_recoverConfig`.
@deprecated(
    "This is a deprecated way of running GanDLF. Please, use `gandlf recover-config` cli command "
    + "instead of `gandlf_recoverConfig`. Note that in new CLI tool some params were renamed or changed its behavior:\n"
    + "  --modeldir to --model-dir\n"
    + "  --mlcube is flag now with default False if not passed. Does not require to pass any additional values\n"
    + "  --outputFile to --output-file`\n"
    + "`gandlf_recoverConfig` script would be deprecated soon."
)
def old_way():
    logger_setup()
    parser = argparse.ArgumentParser(
        prog="GANDLF_RecoverConfig",
        formatter_class=argparse.RawTextHelpFormatter,
        description="Recovers a config file from a GaNDLF model. If used from within a deployed GaNDLF MLCube, attempts to extract the config from the embedded model.\n\n"
        + copyrightMessage,
    )

    parser.add_argument(
        "-m", "--modeldir", metavar="", type=str, help="Path to the model directory."
    )
    # TODO: despite of `str` type, real value is never used (only checks if it is filled or not)
    #  Thus, caveats:
    #    * passing `--mlcube False` would still process it as mlcube;
    #    * passing `--mlcube "" ` (with empty str) acts as non-mlcube
    parser.add_argument(
        "-c",
        "--mlcube",
        metavar="",
        type=str,
        help="Pass this option to attempt to extract the config from the embedded model in a GaNDLF MLCube (if any). Only useful in that context. If passed, model-dir param is ignored.",
    )
    parser.add_argument(
        "-o",
        "--outputFile",
        metavar="",
        type=str,
        required=True,
        help="Path to an output file where the config will be written.",
    )

    args = parser.parse_args()

    _recover_config(args.modeldir, args.mlcube, args.outputFile)


if __name__ == "__main__":
    old_way()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar 11 14:55:44 2021

@author: siddhesh
"""

import os
from typing import Dict, List, Union
import torch


class Logger:
    def __init__(
        self,
        logger_csv_filename: str,
        metrics: List[str],
        mode: str,
        add_epsilon: bool = False,
    ) -> None:
        """
        Logger class to log the training and validation metrics to a csv file. May append to existing file if headers match; elsewise raises an error.

        Args:
            logger_csv_filename (str): Path to a filename where the csv has to be stored.
            metrics (Dict[str, float]): The metrics to be logged.
            mode (str): The mode of the logger, used as suffix to metric names. Normally may be `train` / `val` / `test`
            add_epsilon (bool): Whether to log epsilon values or not (differential privacy measurement)
        """
        self.filename = logger_csv_filename
        mode = mode.lower()
        self.mode = mode.lower()

        new_header = ["epoch_no", f"{mode}_loss"] + [
            f"{mode}_{metric}" for metric in metrics
        ]
        if add_epsilon:
            new_header.append(f"{self.mode}_epsilon")

        # TODO: do we really need to support appending to existing files?
        if os.path.exists(self.filename):
            with open(self.filename, "r") as f:
                existing_header = f.readline().strip().split(",")
            if set(existing_header) != set(new_header):
                raise ValueError(
                    f"Logger file {self.filename} error: existing header does not match new header."
                    f" Existing header: {existing_header}. New header: {new_header}"
                )
            self.ordered_header = existing_header
        else:
            with open(self.filename, "w") as f:
                f.write(",".join(new_header) + "\n")
            self.ordered_header = new_header

    def write(
        self,
        epoch_number: int,
        loss: Union[float, torch.Tensor],
        epoch_metrics: Dict[str, Union[float, torch.Tensor]],
    ) -> None:
        """
        Write the epoch number, loss and metrics to the csv file.

        Args:
            epoch_number (int): The epoch number.
            loss (float): The loss value.
            epoch_metrics (Dict[str, float]): The metrics to be logged.
        """

        if torch.is_tensor(loss):
            loss = loss.cpu().item()

        row = {"epoch_no": epoch_number, f"{self.mode}_loss": loss}

        for metric, metric_val in epoch_metrics.items():
            if torch.is_tensor(metric_val):
                metric_val = metric_val.cpu().item()
            row[f"{self.mode}_{metric}"] = metric_val

        with open(self.filename, "a") as f:
            line = [row.get(col, "") for col in self.ordered_header]
            line = [str(x) for x in line]
            f.write(",".join(line) + "\n")
